package com.stpos.a8pos.devicesdk.impl.cots

import android.content.Context
import android.nfc.NfcAdapter
import android.nfc.NfcAdapter.ReaderCallback
import android.nfc.tech.IsoDep
import com.stpos.a8pos.A8PosApplication
import com.stpos.a8pos.common.handler.StringActionHandler
import com.stpos.a8pos.devicesdk.emv.CardType
import com.stpos.a8pos.devicesdk.emv.EmvTags
import com.stpos.a8pos.devicesdk.emv.KernelID
import com.stpos.a8pos.devicesdk.impl.cots.ppse.PPSEManager
import com.stpos.a8pos.devicesdk.impl.cots.visa.VisaTTPKernelHelper
import com.stpos.a8pos.devicesdk.impl.mock.MockBeeper
import com.stpos.a8pos.devicesdk.impl.mock.emv.IsoDepPosNfcTransceiver
import com.stpos.a8pos.devicesdk.impl.mock.emv.MockEmvData
import com.stpos.a8pos.devicesdk.impl.util.EmvParametersHelper
import com.stpos.a8pos.devicesdk.impl.util.SearchCardInputData
import com.stpos.a8pos.sale.EmvParameterInitializer
import com.stpos.a8pos.service.CardData
import com.stpos.a8pos.service.CardEvent
import com.stpos.a8pos.service.CardReader
import com.stpos.a8pos.service.PreferencesService
import com.stpos.a8pos.transaction.TransactionType
import com.stpos.a8pos.utils.cipher.SymmetricCipher
import com.stpos.a8pos.utils.cipher.TripleDESCDC
import com.util.data.BytesUtil
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.ObservableEmitter
import io.reactivex.disposables.Disposable
import timber.log.Timber
import java.io.IOException
import java.io.UnsupportedEncodingException
import java.math.BigDecimal
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean
import javax.crypto.KeyGenerator


private const val nfcReaderFlags = NfcAdapter.FLAG_READER_NFC_A or NfcAdapter.FLAG_READER_NFC_B or NfcAdapter.FLAG_READER_NFC_F or NfcAdapter.FLAG_READER_NFC_V or NfcAdapter.FLAG_READER_NFC_BARCODE

// list of supported AIDs
private val visaSupportedAids: ArrayList<String> = ArrayList(
    listOf(
        "A0000000031010"
    )
)
private val supportedAids = visaSupportedAids

// default timeout in seconds
private const val readCardTimeoutSeconds = 60L

// TODO proper initialization once details regarding key exchange with COTS terminals have been settled with the acquirer
private val cipherInitialiationVecator = ByteArray(8)
private val keygen: KeyGenerator = KeyGenerator.getInstance("DESede")

// random keys
private val ttkKey = keygen.generateKey().encoded
private val pinKey = keygen.generateKey().encoded

class COTSCardReader(
    private val context: Context,
    private val preferencesService: PreferencesService
) : CardReader {
    private val ttkCipher: SymmetricCipher = TripleDESCDC(ttkKey, cipherInitialiationVecator)
    private val pinCipher: SymmetricCipher = TripleDESCDC(pinKey, cipherInitialiationVecator)

    /**
     * Last search card input data
     */
    private var lastSearchCardInputData: SearchCardInputData? = null

    /**
     * Event emitter object
     */
    private lateinit var eventEmitter: ObservableEmitter<CardEvent>

    /**
     * Mock beeper used to make sound notifications to the cardholder
     */
    private val mockBeeper = MockBeeper()

    /**
     * Holds the reference to a the nfc adapter
     */
    private val nfcAdapter = NfcAdapter.getDefaultAdapter(context)

    /**
     * List of currently supported AIDs (might get updated when an unsupported application is encountered)
     */
    private val currentSupportedAids = mutableListOf<String>()

    /**
     * Reference to the card reader timeout
     */
    private var timeoutDisposable: Disposable? = null

    private fun nfcReadStart(callback: ReaderCallback?) =
        nfcAdapter.enableReaderMode(
            A8PosApplication.mainActivity.get(),
            callback,
            nfcReaderFlags,
            null
        )

    private fun nfcReadStop() =
        nfcAdapter.disableReaderMode(A8PosApplication.mainActivity.get())

    override fun searchForCard(sessionType: TransactionType, amount: BigDecimal?, readCardHandler: StringActionHandler?): Observable<CardEvent> {
        // wrap the parameters so it can be reused when the reading has to be restarted
        val searchCardData = SearchCardInputData(sessionType, amount, readCardHandler)
        lastSearchCardInputData = searchCardData

        return Observable.create {
            eventEmitter = it // store the event emitter
            searchForCardInternal(searchCardData)
        }
    }

    private fun searchForCardInternal(searchCardData: SearchCardInputData) {
        // reset the list of supported AIDs
        currentSupportedAids.clear()
        currentSupportedAids.addAll(supportedAids)

        // card reading has been initiated
        eventEmitter.onNext(CardEvent.ReadCardStarted(readCardTimeoutSeconds))
        startTimeout()

        val ongoingCardRead = AtomicBoolean(false)
        nfcReadStart {
            visaLog("$this - read a tag, ${ongoingCardRead.get()}, tag=${BytesUtil.byte2hex(it.id)} techList=\"${it.techList.joinToString()}\"")
            if (ongoingCardRead.compareAndSet(false, true)) {
                // reset timeout
                startTimeout()
                eventEmitter.onNext(CardEvent.CardRead)
                // notify that a card is being processed
                eventEmitter.onNext(CardEvent.ProcessingCard)
                mockBeeper.startBeep(500)

                try {
                    // initiate ISO dep connection
                    val isoDep = IsoDep.get(it)
                    isoDep.connect()
                    val nfcTransceiver = IsoDepPosNfcTransceiver(isoDep)

                    // init the PPSE manager (to be able to select an application)
                    val ppseManager = PPSEManager()

                    // set supported AIDs
                    ppseManager.setSupportedApps(currentSupportedAids)

                    // initiate application selection
                    val selectedAid = ppseManager.selectPPSE(nfcTransceiver)

                    if (selectedAid != null) {
                        // get transaction config and record based on the search card data
                        val (transactionConfig, transactionRecord) = EmvParametersHelper(preferencesService).getTransactionConfigAndRecordForSearchCardData(searchCardData)

                        val initializer = EmvParameterInitializer(transactionConfig, transactionRecord)
                        val tlvList = initializer.initEmvParameters(
                            String(selectedAid), // AID
                            KernelID.VISA.toByte(), // kernel ID
                            null, // callback
                            CardType.RF, // card type (always RF)
                            MockEmvData(), // mock EMV data
                            searchCardData.sessionType, // transaction type
                            CardData(), // empty card data
                            preferencesService // preferences service
                        )
                            .apply {
                                // add the merchant name to the TLV list
                                val merchantName = preferencesService.merchantName.let { merchantName ->
                                    try {
                                        merchantName.toByteArray(charset("GBK"))
                                    } catch (e: UnsupportedEncodingException) {
                                        merchantName.toByteArray()
                                    }
                                }
                                addTLV(EmvTags.EMV_TAG_TM_MCHNAMELOC, merchantName)
                            }

                        val selectedAidString = BytesUtil.byte2hex(selectedAid)
                        when {
                            // check whether the selected AID is Visa AID
                            visaSupportedAids.contains(selectedAidString) -> {
                                // use the Visa TTP kernel
                                val visaKernelHelper = VisaTTPKernelHelper(preferencesService, nfcTransceiver, pinCipher, mockBeeper)
                                timeoutDisposable?.dispose()
                                visaKernelHelper.performTTPTransaction(
                                    eventEmitter = eventEmitter,
                                    tlvList = tlvList,
                                    searchCardData = searchCardData, selectedAid = selectedAid,
                                    unsupportedAidHandler = { unsupportedAid ->
                                        currentSupportedAids.remove(unsupportedAid)
                                    },
                                    restartSearchingForCard = { restartSearchingForCard() })
                            }
                            else -> {
                                Timber.e("Unsupported AID has been selected")
                                eventEmitter.onNext(CardEvent.Processing)
                                eventEmitter.onNext(CardEvent.CardUnsupportedByTheTerminal {
                                    restartSearchingForCard()
                                })
                            }
                        }
                    } else {
                        // the card read is not supported by the terminal
                        eventEmitter.onNext(CardEvent.Processing)
                        eventEmitter.onNext(CardEvent.CardUnsupportedByTheTerminal {
                            restartSearchingForCard()
                        })
                    }
                } catch (e: IOException) {
                    Timber.e(e, "failed to communicate with a card")
                    eventEmitter.onNext(CardEvent.Processing)
                    eventEmitter.onNext(
                            CardEvent.CardNotAccepted(
                                    reason = CardEvent.CardNotAccepted.Reason.CARD_READ_ERROR,
                                    additionalMessage = null,
                                    tryAgain = {
                                        restartSearchingForCard()
                                    }
                            )
                    )
                } catch (e: NullPointerException) {
                    Timber.e(e, "encountered an NPE")
                    eventEmitter.onNext(CardEvent.Processing)
                    eventEmitter.onNext(CardEvent.CardUnsupportedByTheTerminal {
                        restartSearchingForCard()
                    })
                }
            }
        }
    }

    private fun startTimeout() {
        timeoutDisposable?.dispose()
        // create a timer to cancel the transaction after the given timeout
        timeoutDisposable = Completable.timer(readCardTimeoutSeconds, TimeUnit.SECONDS).subscribe {
            eventEmitter.onNext(CardEvent.TransactionTimeout {
                restartSearchingForCard()
            })
        }
    }

    private fun restartSearchingForCard() {
        lastSearchCardInputData?.let {
            searchForCardInternal(it)
        } ?: run {
            throw IllegalStateException("restartSearchingForCard called but search card input data is null")
        }
    }

    override fun stopEmvAndSearchingForCard() {
        visaLog("stopEmvAndSearchingForCard")
        cleanup()
    }

    override fun restartProcess() {
        visaLog("restartProcess")
        restartSearchingForCard()
    }

    override fun forcePin() {
        visaLog("forcePin")
        // nothing at the moment
    }

    override fun cleanup() {
        visaLog("cleanup")
        timeoutDisposable?.dispose()
        lastSearchCardInputData = null
        nfcReadStop()
    }

}
