package com.stpos.a8pos.devicesdk.impl.cots.visa

import com.stpos.a8pos.A8PosApplication
import com.stpos.a8pos.acquirer.monetplus.MonetAcquirerRequestDol
import com.stpos.a8pos.devicesdk.emv.EmvTags
import com.stpos.a8pos.devicesdk.impl.cots.visaLog
import com.stpos.a8pos.devicesdk.impl.mock.MockBeeper
import com.stpos.a8pos.devicesdk.impl.mock.emv.IsoDepPosNfcTransceiver
import com.stpos.a8pos.devicesdk.impl.util.SearchCardInputData
import com.stpos.a8pos.devicesdk.impl.util.getExpiryDateFromTrack2
import com.stpos.a8pos.emv.TlvProvider
import com.stpos.a8pos.emv.toEmvValue
import com.stpos.a8pos.enums.CardPresentation
import com.stpos.a8pos.service.CardData
import com.stpos.a8pos.service.CardEvent
import com.stpos.a8pos.service.PreferencesService
import com.stpos.a8pos.utils.EmvUtils
import com.stpos.a8pos.utils.cipher.SymmetricCipher
import com.util.data.BytesUtil
import com.util.data.TLV
import com.util.data.TLVList
import com.visa.app.ttpkernel.*
import io.reactivex.Observable
import io.reactivex.ObservableEmitter
import io.reactivex.schedulers.Schedulers
import timber.log.Timber
import java.math.BigDecimal
import java.security.InvalidParameterException
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean
import kotlin.experimental.or


// B1: mkey entry, magnetic stripe, chip card not supported
// B2: online PIN supported
// B3: DDA and CDA supported
private const val cotsVisaTerminalCapabilities = "004048"

// B1: only payment is supported
// B2: cash deposit not supported
// B3: numeric and alphabetic keys supported
private const val cotsVisaAdditionalTerminalCapabilities = "00C80300"

// tag specific to the Visa Kernel SDK namespace
private const val readerCvmRequiredLimitTag = "DF01"

class VisaTTPKernelHelper(
    private val preferencesService: PreferencesService,
    private val nfcTransceiver: IsoDepPosNfcTransceiver,
    private val pinCipher: SymmetricCipher,
    private val mockBeeper: MockBeeper
) {
    fun performTTPTransaction(
        eventEmitter: ObservableEmitter<CardEvent>,
        tlvList: TLVList,
        searchCardData: SearchCardInputData,
        selectedAid: ByteArray,
        unsupportedAidHandler: (String) -> Unit,
        restartSearchingForCard: () -> Unit
    ) {
        // prepare input parameters
        val contactlessKernel: ContactlessKernel = ContactlessKernel.getInstance(A8PosApplication.appContext)
        val contactlessConfiguration = getContactlessConfiguration(tlvList, searchCardData, selectedAid).apply {
            terminalData.forEach { entry ->
                visaLog("\"${entry.key}\": ${BytesUtil.byte2hex(entry.value)}")
            }
        }

        A8PosApplication.mainActivity.get()!!.runOnUiThread {
            eventEmitter.onNext(CardEvent.Processing)
            // perform the transaction
            val kernelResult = contactlessKernel.performTransaction(nfcTransceiver, contactlessConfiguration)
            // Check the transaction outcome
            processKernelResult(kernelResult, selectedAid, eventEmitter, unsupportedAidHandler, restartSearchingForCard)
        }
    }

    private fun getContactlessConfiguration(tlvList: TLVList, searchCardData: SearchCardInputData, selectedAid: ByteArray) = ContactlessConfiguration.getInstance().apply {
        terminalData.apply {
            tlvList
                .asList()
                .forEach { tlv ->
                    // store input TLVs to be able to pass them to the kernel
                    put(tlv.tag, tlv.bytesValue)

                    visaLog("generated \"${tlv.tag}\": ${BytesUtil.byte2hex(tlv.bytesValue)}")
                }

            // this tag (transaction type) is not present above and has to be set
            put(
                EmvTags.EMV_TAG_TM_TRANSTYPE, byteArrayOf(
                    searchCardData.sessionType.toEmvValue()
                        ?: throw InvalidParameterException("Only Payment and Refund transactions are supported")
                )
            )

            // add missing TLVs (terminal capabilities and some TTP kernel specific tags)
            put(EmvTags.EMV_TAG_TM_CAP, BytesUtil.hexString2ByteArray(cotsVisaTerminalCapabilities))
            put(EmvTags.EMV_TAG_TM_CAP_AD, BytesUtil.hexString2ByteArray(cotsVisaAdditionalTerminalCapabilities))
            put(EmvTags.C_TAG_TM_DF69, BytesUtil.hexString2ByteArray("00"))
            put(EmvTags.DEF_TAG_GAC_CONTROL, BytesUtil.hexString2ByteArray("00"))

            // set CVM required byte based on whether the transaction amount is above the CVM limit
            val cvmRequiredByte = if (searchCardData.amount!! >= BigDecimal.valueOf(preferencesService.transLimitNoPinCless).movePointLeft(2)) {
                0x40.toByte()
            } else {
                0x00.toByte()
            }

            // TTQ must contain the Online PIN capability
            put(EmvTags.V_TAG_TM_9F66, byteArrayOf(0x24.toByte(), 0x80.toByte() or cvmRequiredByte, 0x40.toByte(), 0x00.toByte()))
            put(EmvTags.EMV_TAG_TM_POSENTMODE, byteArrayOf(0x07)) //POS Entry Mode
            put(EmvTags.EMV_TAG_IC_AID, selectedAid) // set the selected aid)

            // CVM limit specific to TTP Kernel namespace
            put(
                readerCvmRequiredLimitTag, BytesUtil.toBCDAmountBytes(
                    BigDecimal(preferencesService.transLimitNoPinCless).toLong()
                )
            )
        }
    }

    /**
     * Processes the given kernel result.
     *
     * @param kernelResult kernel result to be processed
     * @param selectedAid the AID that was selected during AID selection
     */
    private fun processKernelResult(
        kernelResult: ContactlessResult,
        selectedAid: ByteArray,
        eventEmitter: ObservableEmitter<CardEvent>,
        unsupportedAidHandler: (String) -> Unit,
        restartSearchingForCard: () -> Unit,

        ) {
        if (kernelResult.lastSW == null) {
            // when last SW (command result) is null, an error is indicated
            visaLog("kernelResult.lastSW is null, restarted searching for card")
            eventEmitter.onNext(
                    CardEvent.CardNotAccepted(
                            reason = CardEvent.CardNotAccepted.Reason.CARD_READ_ERROR,
                            additionalMessage = null,
                            tryAgain = {
                                restartSearchingForCard()
                            }
                    )
            )
            return
        }

        // log the kernel result
        visaLog("kernelResult.lastSW: ${BytesUtil.byte2hex(kernelResult.lastSW)}")
        visaLog("kernelResult.finalOutcome: ${kernelResult.finalOutcome.name}")
        visaLog("""kernelResult.data: 
            ${
            kernelResult.data.entries.joinToString(separator = "\n") { (tagName, value) ->
                "$tagName => ${value?.let { BytesUtil.byte2hex(it) } ?: "null"}"
            }
        }""".trimIndent())
        visaLog("""kernelResult.internalData: 
            ${
            kernelResult.internalData.entries.joinToString(separator = "\n") { (tagName, value) ->
                "$tagName => ${value?.let { BytesUtil.byte2hex(it) } ?: "null"}"
            }
        }""".trimIndent())
        visaLog("kernelResult.finalOutcome: ${kernelResult.finalOutcome.name}")

        fun getCardData(kernelTlvProvider: TlvProvider, onlinePin: String?) = CardData().apply {
            presentationMethod = CardPresentation.CLESS

            // get request data
            val acquirerDol = MonetAcquirerRequestDol(kernelTlvProvider)
            emvRequestData = "01" + acquirerDol.getRequestData()
            additionalEmvRequestData = "02" + acquirerDol.getAdditionalRequestData()

            // add the PIN to the card data  if provided
            if (onlinePin != null) {
                addFlag(CardData.CardFlag.PIN_ENTERED)
                pinBlock = pinCipher.encrypt(BytesUtil.ascii2Bcd(onlinePin))
            }

            // get track2 and card number from TLV tags
            track2 = kernelTlvProvider.getTlv(EmvTags.EMV_TAG_IC_TRACK2DATA)!!.value.let {
                // PAN from Visa TTP is padded with 4 characters
                if (it.length >= 16) {
                    it.substring(4, it.length)
                } else {
                    it
                }
            }

            // extract the card number (PAN) and the expiration date from the list of tags
            cardNumber = kernelTlvProvider.getTlv(EmvTags.EMV_TAG_IC_PAN)?.value
                ?: EmvUtils.getPanFromTrack2(EmvUtils.normalizeTrack2(track2))

            expirationDate = kernelTlvProvider.getTlv(EmvTags.EMV_TAG_IC_APPEXPIREDATE)?.value
                ?: getExpiryDateFromTrack2(track2!!)
        }

        when (kernelResult.finalOutcome) {
            TtpOutcome.COMPLETED -> {
                Timber.d("VisaTTP transaction completed")
                val cvmOutcome = TtpCVMKernelOutcome.values()[kernelResult.internalData["DF03"]?.get(3)?.toInt()!!]
                visaLog("kernelResult.cvmOutcome: $cvmOutcome")
                when (cvmOutcome) {
                    TtpCVMKernelOutcome.CVM_NOT_REQUIRED -> {
                        Timber.d("omitting online request event")
                        eventEmitter.onNext(
                            CardEvent.OnlineRequest(
                                data = getCardData(
                                    kernelTlvProvider = getTlvProvider(kernelResult.data),
                                    onlinePin = null
                                ),
                                callback = { _, _ ->
                                    // TODO online authorization result can't be validated by Visa TTP right now, we must trust the result's authenticity
                                }
                            )
                        )
                    }
                    // signature is not supported right now
                    TtpCVMKernelOutcome.SIGNATURE -> throw IllegalStateException("SIGNATURE not supported")
                    TtpCVMKernelOutcome.ONLINE_PIN -> {
                        // Online PIN CVM required
                        val pinSubmitted = AtomicBoolean(false)

                        // prompt for PIN using the SW PIN pad
                        eventEmitter.onNext(CardEvent.PromptForSoftPin(object : CardEvent.PromptForSoftPin.PinPadResult {
                            override fun pinSubmitted(pin: String) {
                                // pin has been submitted
                                pinSubmitted.set(true)
                                visaLog("pinSubmitted")
                                eventEmitter.onNext(CardEvent.PinEnteringFinished)
                                eventEmitter.onNext(CardEvent.OnlineRequest(
                                    data = getCardData(getTlvProvider(kernelResult.data), pin),
                                    callback = { _, _ ->
                                        // TODO online authorization result can't be validated by Visa TTP right now, we must trust the result's authenticity
                                    }
                                ))
                            }

                            override fun pinCancelled() {
                                visaLog("pinCancelled")
                                if (!pinSubmitted.get()) {
                                    eventEmitter.onNext(CardEvent.PinEnteringCanceled)
                                    eventEmitter.onNext(CardEvent.PinEnteringFinished)
                                }
                            }

                            override fun keyDown() {
                                visaLog("keyDown")
                                mockBeeper.startBeep(150)
                            }

                        }))
                    }
                    TtpCVMKernelOutcome.CVM_DECLINED -> {
                        // the transaction has been declined due to a failed CVM
                        eventEmitter.onNext(
                            CardEvent.CardNotAccepted(
                                reason = CardEvent.CardNotAccepted.Reason.EMV_DECLINED,
                                additionalMessage = null,
                                tryAgain = {
                                    restartSearchingForCard()
                                }
                            )
                        )
                    }
                }

            }
            TtpOutcome.DECLINED -> {
                // declined by the kernel
                Timber.d("VisaTTP transaction declined")
                eventEmitter.onNext(
                    CardEvent.CardNotAccepted(
                        reason = CardEvent.CardNotAccepted.Reason.EMV_DECLINED,
                        additionalMessage = null,
                        tryAgain = {
                            restartSearchingForCard()
                        }
                    ))
            }
            TtpOutcome.ABORTED -> {
                Timber.d("VisaTTP transaction aborted")
                eventEmitter.onNext(
                    CardEvent.CardNotAccepted(
                        reason = CardEvent.CardNotAccepted.Reason.CARD_READ_ERROR,
                        additionalMessage = null,
                        tryAgain = {
                            restartSearchingForCard()
                        }
                    )
                )
            }
            TtpOutcome.TRYNEXT -> {
                Timber.d("VisaTTP transaction aborted")
                unsupportedAidHandler.invoke(String(selectedAid))
                eventEmitter.onNext(
                    CardEvent.RepresentCard(
                        reason = CardEvent.RepresentCard.Reason.APDU_ERROR,
                        callback = { restartSearchingForCard() }
                    )
                )
            }
            TtpOutcome.SELECTAGAIN -> {
                // this outcome is emitted when CDCVM is eneded
                eventEmitter.onNext(CardEvent.CdCvmNeeded {
                    // nothing (is invoked when user taps the OK button)
                })
                // restart searching for cards after 2 seconds
                Observable.timer(2, TimeUnit.SECONDS)
                    .subscribeOn(Schedulers.newThread())
                    .subscribe {
                        restartSearchingForCard()
                    }
            }
            else -> {
                eventEmitter.onNext(CardEvent.GeneralError {
                    restartSearchingForCard()
                })
            }
        }
    }


    private fun getTlvProvider(data: HashMap<String, ByteArray>) = object : TlvProvider {
        override fun getTlv(tag: String): TLV? =
            data.get(tag)?.let {
                TLV.fromData(tag, it)
            }
    }

}