package com.stpos.a8pos.devicesdk.impl.nexgo

import com.nexgo.oaf.apiv3.DeviceEngine
import com.nexgo.oaf.apiv3.SdkResult
import com.nexgo.oaf.apiv3.device.led.LightModeEnum
import com.nexgo.oaf.apiv3.device.reader.CardInfoEntity
import com.nexgo.oaf.apiv3.device.reader.CardSlotTypeEnum
import com.stpos.a8pos.common.handler.StringActionHandler
import com.stpos.a8pos.devicesdk.impl.nexgo.card.NexgoCardReaderHelper
import com.stpos.a8pos.devicesdk.impl.nexgo.card.NexgoCardReaderHelper.SearchError.*
import com.stpos.a8pos.devicesdk.impl.nexgo.card.NexgoCardReaderHelperImpl
import com.stpos.a8pos.devicesdk.impl.nexgo.emv.NexgoEmvProcessHelper
import com.stpos.a8pos.devicesdk.impl.nexgo.emv.NexgoEmvInitializer
import com.stpos.a8pos.devicesdk.impl.nexgo.emv.NexgoEmvProcessHelperImpl
import com.stpos.a8pos.devicesdk.impl.nexgo.emv.NexgoEmvInitializerImpl
import com.stpos.a8pos.devicesdk.impl.util.SearchCardInputData
import com.stpos.a8pos.devicesdk.impl.nexgo.util.nexgoDebugLog
import com.stpos.a8pos.service.CardEvent
import com.stpos.a8pos.service.CardReader
import com.stpos.a8pos.service.KeyIdService
import com.stpos.a8pos.service.PreferencesService
import com.stpos.a8pos.transaction.TransactionType
import io.reactivex.Observable
import io.reactivex.ObservableEmitter
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import timber.log.Timber
import java.math.BigDecimal
import java.util.concurrent.TimeUnit


// support only standard contact and contactless interfaces
private val supportedInterfaces = setOf(CardSlotTypeEnum.ICC1, CardSlotTypeEnum.RF)

/**
 * CardReader implementation to be used on Nexgo devices.
 */
class NexgoCardReader(
    private val apiEngine: DeviceEngine,
    private val preferencesService: PreferencesService,
    private val keyIdService: KeyIdService
) : CardReader,
    NexgoEmvInitializer by NexgoEmvInitializerImpl,
    NexgoCardReaderHelper by NexgoCardReaderHelperImpl(
        cardReader = apiEngine.cardReader,
        cardReadTimeoutMs = 60000
    ),
    NexgoEmvProcessHelper by NexgoEmvProcessHelperImpl(
        cardReader = apiEngine.cardReader,
        pinpad = apiEngine.pinPad,
        led = apiEngine.ledDriver,
        keyIdService,
        emvHandler = apiEngine.getEmvHandler("app1"),
        preferencesService = preferencesService
    ) {

    /**
     * Last presented card information
     */
    private var lastCardInfoEntity: CardInfoEntity? = null

    /**
     * Event output stream
     */
    private lateinit var eventEmitter: ObservableEmitter<CardEvent>

    /**
     * Device LED driver
     */
    private val led = apiEngine.ledDriver

    /**
     * Device card reader
     */
    private val cardReader = apiEngine.cardReader

    /**
     * Device EMV process handler
     */
    private val emvHandler = apiEngine.getEmvHandler("app1")

    /**
     * Search card request input data (utilized when restarting the card search process)
     */
    private var currentSearchCardInputData: SearchCardInputData? = null

    /**
     * Used to hold the reference to the disposable object
     * that restarts the search for card process.
     */
    private var restartDisposable: Disposable? = null

    override fun searchForCard(sessionType: TransactionType, amount: BigDecimal?, readCardHandler: StringActionHandler?): Observable<CardEvent> {
        val searchCardData = SearchCardInputData(sessionType, amount, readCardHandler)
        currentSearchCardInputData = searchCardData

        nexgoDebugLog("searchForCard")
        return Observable.create {
            eventEmitter = it
            searchForCardInternal(searchCardData)
        }
    }

    private fun searchForCardInternal(searchCardInputData: SearchCardInputData) {
        // reset current data
        resetSessionData()

        // initialize the EMV process
        val (capkInitResult, aidParaInitResult) = initializeEmv(emvHandler)

        if (capkInitResult != SdkResult.Success && aidParaInitResult != SdkResult.Success) {
            Timber.e("Could not initialize CAPK and AID lists, capkInitResult=$capkInitResult, aidParaInitResult=$aidParaInitResult")
            stopEmvAndSearchingForCard()
            eventEmitter.onNext(CardEvent.GeneralError { })
            return
        }

        // set up device's LED
        led.setLed(LightModeEnum.BLUE, true)

        // notify that card reading has started
        eventEmitter.onNext(CardEvent.ReadCardStarted(60))

        // search for card
        val readStarted = searchCard(
            interfaces = supportedInterfaces,
            onCardRead = {
                onSearchCardSuccess(it, searchCardInputData)
            },
            onError = {
                led.setLed(LightModeEnum.BLUE, false)
                when (it) {
                    INVALID_SWIPE -> eventEmitter.onNext(CardEvent.SwipeError)
                    MULTIPLE_CARDS -> eventEmitter.onNext(CardEvent.CardCollision(true))
                    TIMEOUT -> eventEmitter.onNext(CardEvent.TransactionTimeout {
                        restartSearchForCard()
                    })
                    FAIL -> eventEmitter.onNext(CardEvent.GeneralError {})
                }
            }
        )

        if (!readStarted) {
            Timber.e("Could not start searching for the card")
            cleanup()
            eventEmitter.onNext(CardEvent.GeneralError {})
        }

    }

    private fun resetSessionData() {
        nexgoDebugLog("resetSessionData")
        lastCardInfoEntity = null
    }

    private fun onSearchCardSuccess(cardInfoEntity: CardInfoEntity, searchCardInputData: SearchCardInputData) {
        eventEmitter.onNext(CardEvent.ProcessingCard)
        led.setLed(LightModeEnum.BLUE, false)
        lastCardInfoEntity = cardInfoEntity
        // card found, start the emv process
        startEmvProcess(
            initCardInfoEntity = cardInfoEntity,
            searchCardInputData = searchCardInputData,
            eventEmitter = eventEmitter,
            restartSearchForCard = { restartSearchForCard() },
            cleanup = { cleanup() },
            stopEmvAndSearchingForCard = {
                stopEmvAndSearchingForCard()
            },
            restartSearchForCardDelayed = { restartSearchForCardDelayed() }
        )
    }

    private fun restartSearchForCard() {
        nexgoDebugLog("restartSearchForCard")
        currentSearchCardInputData?.let {
            searchForCardInternal(it)
        } ?: run {
            throw IllegalStateException("restartSearchForCard was called but currentSearchCardInputData is not set")
        }
    }

    override fun stopEmvAndSearchingForCard() {
        nexgoDebugLog("stopEmvAndSearchingForCard")
        cardReader.stopSearch()
        emvHandler.emvProcessCancel()
    }

    override fun restartProcess() {
        TODO("Not yet implemented")
    }

    override fun forcePin() {
        TODO("Not yet implemented")
    }

    override fun cleanup() {
        restartDisposable?.dispose()
        nexgoDebugLog("cleanup")
        stopEmvAndSearchingForCard()
        resetSessionData()
        currentSearchCardInputData = null
        led.setLed(LightModeEnum.BLUE, false)
    }

    private fun restartSearchForCardDelayed() {
        cardReader.stopSearch()
        restartDisposable = Observable.timer(3, TimeUnit.SECONDS)
            .subscribeOn(Schedulers.newThread()) // observe on MT?
            .subscribe {
                restartSearchForCard()
            }
    }
}