package com.stpos.a8pos.devicesdk.impl.nexgo.card

import com.nexgo.oaf.apiv3.SdkResult
import com.nexgo.oaf.apiv3.device.reader.CardInfoEntity
import com.nexgo.oaf.apiv3.device.reader.CardReader
import com.nexgo.oaf.apiv3.device.reader.CardSlotTypeEnum
import com.nexgo.oaf.apiv3.device.reader.OnCardInfoListener
import com.stpos.a8pos.devicesdk.impl.nexgo.util.nexgoDebugLog

interface NexgoCardReaderHelper {

    enum class SearchError {
        INVALID_SWIPE,
        MULTIPLE_CARDS,
        TIMEOUT,
        FAIL
    }

    /**
     * Searches for card in given interfaces.
     * Notifies the passed listeners based on the search progress.
     *
     * @param interfaces set of interfaces in which cards should be searched for
     * @param onCardRead a function to be invoked when a card has been successfully read
     * @param onError a function to be invoked when an error has occurred during
     *
     * @return a boolean flag indicating whether the search process has been successfully started.
     */
    fun searchCard(
            interfaces: Set<CardSlotTypeEnum>,
            onCardRead: (CardInfoEntity) -> Unit,
            onError: (SearchError) -> Unit): Boolean
}

class NexgoCardReaderHelperImpl(
        private val cardReader: CardReader,
        private val cardReadTimeoutMs: Int,
        ) : NexgoCardReaderHelper {
    override fun searchCard(
            interfaces: Set<CardSlotTypeEnum>,
            onCardRead: (CardInfoEntity) -> Unit,
            onError: (NexgoCardReaderHelper.SearchError) -> Unit): Boolean {
        // initiate searching for card using the CardReader class from the Nexgo SDK
        val result = cardReader.searchCard(
                interfaces.toHashSet(),
                cardReadTimeoutMs / 1000,
                object : OnCardInfoListener {
                    override fun onCardInfo(retCode: Int, cardInfoEntity: CardInfoEntity?) {
                        nexgoDebugLog("onCardInfo retCode: $retCode")
                        when (retCode) {
                            SdkResult.Success -> onCardRead(cardInfoEntity!!)
                            SdkResult.Fail -> onError(NexgoCardReaderHelper.SearchError.FAIL)
                            SdkResult.TimeOut -> onError(NexgoCardReaderHelper.SearchError.TIMEOUT)
                        }
                    }

                    override fun onSwipeIncorrect() =
                            onError(NexgoCardReaderHelper.SearchError.INVALID_SWIPE)

                    override fun onMultipleCards() =
                            onError(NexgoCardReaderHelper.SearchError.MULTIPLE_CARDS)

                })

        return result == SdkResult.Success
    }
}