package com.stpos.a8pos.devicesdk.impl.nexgo.emv

import com.nexgo.common.ByteUtils
import com.nexgo.oaf.apiv3.SdkResult
import com.nexgo.oaf.apiv3.device.led.LEDDriver
import com.nexgo.oaf.apiv3.device.led.LightModeEnum
import com.nexgo.oaf.apiv3.device.pinpad.*
import com.nexgo.oaf.apiv3.device.reader.CardInfoEntity
import com.nexgo.oaf.apiv3.device.reader.CardReader
import com.nexgo.oaf.apiv3.device.reader.CardSlotTypeEnum
import com.nexgo.oaf.apiv3.emv.*
import com.stpos.a8pos.devicesdk.emv.CandidateAID
import com.stpos.a8pos.devicesdk.emv.CardType
import com.stpos.a8pos.devicesdk.emv.EmvTags
import com.stpos.a8pos.devicesdk.emv.KernelID
import com.stpos.a8pos.devicesdk.impl.util.EmvParametersHelper
import com.stpos.a8pos.devicesdk.impl.util.SearchCardInputData
import com.stpos.a8pos.devicesdk.impl.util.getExpiryDateFromTrack2
import com.stpos.a8pos.devicesdk.impl.nexgo.util.nexgoDebugLog
import com.stpos.a8pos.devicesdk.impl.nexgo.util.nexgoErrorLog
import com.stpos.a8pos.enums.CardPresentation
import com.stpos.a8pos.sale.EmvParameterInitializer
import com.stpos.a8pos.service.CardData
import com.stpos.a8pos.service.CardEvent
import com.stpos.a8pos.service.KeyIdService
import com.stpos.a8pos.service.PreferencesService
import com.stpos.a8pos.transaction.TransactionType
import com.util.data.BytesUtil
import com.util.data.TLV
import com.util.data.TLVList
import io.reactivex.ObservableEmitter
import timber.log.Timber
import java.io.UnsupportedEncodingException
import java.lang.Exception
import java.math.BigDecimal
import java.security.InvalidParameterException
import java.text.SimpleDateFormat
import java.util.*

private val emvRequestDataTags = arrayOf(
        TlvTags.cryptogramInformationData,
        TlvTags.terminalCountryCode,
        TlvTags.transactionDate,
        TlvTags.applicationCryptogram,
        TlvTags.applicationInterchangeProfile,
        TlvTags.applicationTransactionCounter,
        TlvTags.unpredictableNumber,
        TlvTags.terminalVerificationResults,
        TlvTags.transactionType,
        TlvTags.transactionCurrencyCode,
        TlvTags.transactionAmount,
        TlvTags.issuerApplicationData
)

private val emvRequestData2Tags = arrayOf(
        TlvTags.panSequenceNumber,
        TlvTags.terminalType,
        TlvTags.cvmResults,
        TlvTags.applicationVersionNumber,
        TlvTags.terminalCapabilities,
        TlvTags.dedicatedFileName
)

/**
 * Used to initialize an EMV dialogue with the found card.
 */
interface NexgoEmvProcessHelper {
    fun startEmvProcess(
            initCardInfoEntity: CardInfoEntity,
            searchCardInputData: SearchCardInputData,
            eventEmitter: ObservableEmitter<CardEvent>,
            restartSearchForCard: () -> Unit,
            cleanup: () -> Unit,
            stopEmvAndSearchingForCard: () -> Unit,
            restartSearchForCardDelayed: () -> Unit
    )
}

class NexgoEmvProcessHelperImpl(
        private val cardReader: CardReader,
        private val pinpad: PinPad,
        private val led: LEDDriver,
        private val keyIdService: KeyIdService,
        private val emvHandler: EmvHandler,
        private val preferencesService: PreferencesService) : NexgoEmvProcessHelper {

    /**
     * A flag indicating whether incorrect PIN was just entered during offline CVM
     */
    private var incorrectOfflinePinJustEntered = false

    /**
     * The initial CardInfoEntity object that was used to initiate the EMV process
     */
    private var initCardInfoEntity: CardInfoEntity? = null

    /**
     * The initial data used to initialize the transaction
     */
    private var searchCardInputData: SearchCardInputData? = null

    /**
     * Byte array containing the encrypted PIN block that might have been entered by the cardholder
     */
    private var submittedPinBlock: ByteArray? = null

    /**
     * Flag indicating whether online PIN has been submitted
     */
    private var onlinePinRequested = false

    private fun getEmvTransData(cardInfoEntity: CardInfoEntity, searchCardInputData: SearchCardInputData) = EmvTransDataEntity().apply {
        // set transaction type
        b9C = when (searchCardInputData.sessionType) {
            TransactionType.PAYMENT -> 0x00.toByte()
            TransactionType.REFUND -> 0x20.toByte()
            else -> throw InvalidParameterException("Other than PAYMENT or REFUND transactions are not supported. Received: ${searchCardInputData.sessionType}")
        }

        // set transaction amount
        transAmt = (searchCardInputData.amount
                ?: BigDecimal.ZERO).movePointRight(2).toLong().toString()

        // set terminal and merchant IDs
        termId = preferencesService.terminalNumber
        merId = preferencesService.merchantNumber

        // set transaction time
        Date().also {
            transDate = SimpleDateFormat("yyMMdd", Locale.getDefault()).format(it)
            transTime = SimpleDateFormat("hhmmss", Locale.getDefault()).format(it)
        }

        // TODO unknown trace number
        traceNo = "00000000"

        // CDCVM supported
        setIsSupportCDCVM(true)

        // set transaction flow
        procType = EmvTransFlowEnum.FULL
        setIsForceOnline(true)

        // determine the channel type
        channelType = if (cardInfoEntity.cardExistslot == CardSlotTypeEnum.RF) {
            EmvChannelTypeEnum.FROM_PICC
        } else {
            EmvChannelTypeEnum.FROM_ICC
        }
    }

    // the initial TLV list should contain the country code and currency code the merchant is using
    private fun getInitTlvList() = TLVList().apply {
        addTLV(TlvTags.terminalCountryCode, ByteUtils.commonString2Bcd(preferencesService.countryCode))
        addTLV(TlvTags.transactionCurrencyCode, ByteUtils.commonString2Bcd(preferencesService.currencyCode))
        addTLV(TlvTags.transactionReferenceCurrencyCode, ByteUtils.commonString2Bcd(preferencesService.currencyCode))
    }

    override fun startEmvProcess(
            initCardInfoEntity: CardInfoEntity,
            searchCardInputData: SearchCardInputData,
            eventEmitter: ObservableEmitter<CardEvent>,
            restartSearchForCard: () -> Unit,
            cleanup: () -> Unit,
            stopEmvAndSearchingForCard: () -> Unit,
            restartSearchForCardDelayed: () -> Unit
    ) {
        this.initCardInfoEntity = initCardInfoEntity
        this.searchCardInputData = searchCardInputData
        this.submittedPinBlock = null
        this.onlinePinRequested = false

        // initialize EMV process
        emvHandler.emvDebugLog(true)

        // initialize EMV with terminal country code and transaction currency code
        emvHandler.initTermConfig(getInitTlvList().toBinary())

        // get EMV transaction data
        val transData = getEmvTransData(initCardInfoEntity, searchCardInputData)

        // start the emv process
        val result = emvHandler.emvProcess(
                transData,
                NexgoEmvProcessListener(
                        usedCardSlotType = initCardInfoEntity.cardExistslot,
                        eventEmitter = eventEmitter,
                        stopEmvAndSearchingForCard = stopEmvAndSearchingForCard,
                        restartSearchForCard = restartSearchForCard,
                        cleanup = cleanup,
                        restartSearchForCardDelayed = restartSearchForCardDelayed
                )
        )

        if (result != SdkResult.Success) {
            nexgoErrorLog("emvProcess returned $result")
        }
    }

    /**
     * An inner class that implements the `OnEmvProcessListener` interface from the Nexgo SDK
     */
    inner class NexgoEmvProcessListener(
            private val usedCardSlotType: CardSlotTypeEnum,
            private val eventEmitter: ObservableEmitter<CardEvent>,
            private val stopEmvAndSearchingForCard: () -> Unit,
            private val restartSearchForCard: () -> Unit,
            private val cleanup: () -> Unit,
            private val restartSearchForCardDelayed: () -> Unit) : OnEmvProcessListener {

        private var confirmedCardInfoEntity: CardInfoEntity? = null

        override fun onConfirmCardNo(cardInfoEntity: CardInfoEntity?) {
            // card info entity called, however, on some AIDs, such as Mastercard, this is not called
            nexgoDebugLog("onConfirmCardNo")
            fun mask(value: String?) = value?.let {
                value.replace(value.substring(4..(value.length - 4)), "#...#")
            } ?: "null"

            confirmedCardInfoEntity = cardInfoEntity

            nexgoDebugLog(
                    """
                            SdkResult.SUCCESS ->
                            cardNo: ${mask(cardInfoEntity?.cardNo)}
                            cardExistslot: ${cardInfoEntity?.cardExistslot}
                            rfCardType: ${cardInfoEntity?.rfCardType}
                            tk1: ${mask(cardInfoEntity?.tk1)}
                            tk2: ${mask(cardInfoEntity?.tk2)}
                            tk3: ${mask(cardInfoEntity?.tk3)}
                            expiredDate: ${cardInfoEntity?.expiredDate}
                            serviceCode: ${cardInfoEntity?.serviceCode}
                            isTk1Valid: ${cardInfoEntity?.isTk1Valid}
                            isTk2Valid: ${cardInfoEntity?.isTk2Valid}
                            isTk3Valid: ${cardInfoEntity?.isTk3Valid}
                            isICC: ${cardInfoEntity?.isICC}
                            """.trimIndent()
            )
            eventEmitter.onNext(CardEvent.CardRead)
            emvHandler.onSetConfirmCardNoResponse(true)
        }

        override fun onCardHolderInputPin(isOnlinePin: Boolean, leftTimes: Int) {
            // the cardholder must input his PIN

            // declare onSendKey handler
            fun handleOnSendKey(keyCode: Byte) {
                nexgoDebugLog("handleOnSendKey: $keyCode")
            }

            // declare pinInputResult handler
            fun handlePinInputResult(retCode: Int, pinData: ByteArray?, online: Boolean) {
                nexgoDebugLog("handlePinInput: $retCode, $pinData, $online, isCardPresent=${cardReader.isCardExist(CardSlotTypeEnum.ICC1)}")

                eventEmitter.onNext(CardEvent.PinEnteringFinished)
                val isCardPresentInReader = cardReader.isCardExist(CardSlotTypeEnum.ICC1)
                if (online || isCardPresentInReader) {
                    // PIN input result should not be handled when the card is
                    // not present in the reader and a chip transaction
                    // is in progress
                    nexgoDebugLog("handlePinInput: $retCode, ${BytesUtil.byte2hex(pinData)}, $online, isCardPresent=${cardReader.isCardExist(CardSlotTypeEnum.ICC1)}")
                    val decrypted = pinpad.calcByWKey(keyIdService.mainKeyIndex, WorkKeyTypeEnum.PINKEY, pinData, pinData!!.size, CalcModeEnum.DECRYPT)
                    nexgoDebugLog("decrypted=${BytesUtil.byte2hex(decrypted)}")
                    nexgoDebugLog("pan=${cardPan}")
                    nexgoDebugLog("pan=${BytesUtil.string2HexString(cardPan)}")

                    // method signature without obfuscation:
                    // emvHandler.onSetPinInputResponse(isConfirm, isBypass)
                    when (retCode) {
                        SdkResult.Success -> {
                            submittedPinBlock = pinData
                            eventEmitter.onNext(CardEvent.PinEnteringFinished)
                            emvHandler.onSetPinInputResponse(true, false)
                        }
                        SdkResult.Fail -> {
                            eventEmitter.onNext(CardEvent.PinError(CardEvent.PinError.Reason.GENERAL_ERROR, "General PIN Failure"))
                            emvHandler.onSetPinInputResponse(false, false)
                        }
                        SdkResult.PinPad_Input_Timeout -> {
                            eventEmitter.onNext(CardEvent.PinError(CardEvent.PinError.Reason.TIMEOUT, "General PIN Failure"))
                            emvHandler.onSetPinInputResponse(false, true)
                        }
                        SdkResult.PinPad_Input_Cancel, SdkResult.PinPad_No_Pin_Input -> {
                            eventEmitter.onNext(CardEvent.Processing)
                            eventEmitter.onNext(CardEvent.PinEnteringCanceled)
                            emvHandler.onSetPinInputResponse(false, true)
                        }
                        else -> {
                            eventEmitter.onNext(CardEvent.PinError(CardEvent.PinError.Reason.GENERAL_ERROR, "Unknown PIN failure"))
                            emvHandler.onSetPinInputResponse(false, true)
                        }
                    }
                } else {
                    stopEmvAndSearchingForCard()
                    eventEmitter.onNext(CardEvent.PinEnteringFinished)
                    eventEmitter.onNext(CardEvent.CardRemovedBeforePinEntered {
                        restartSearchForCard()
                    })
                }
            }

            // declare a PinPad result listener
            fun getPinListener(online: Boolean) = object : OnPinPadInputListener {
                override fun onInputResult(retCode: Int, data: ByteArray?) {
                    handlePinInputResult(retCode, data, online)
                }

                override fun onSendKey(keyCode: Byte) {
                    handleOnSendKey(keyCode)
                }
            }

            // initialize the pin pad
            nexgoDebugLog("onCardHolderInputPin $isOnlinePin, $leftTimes")
            val pinPadInitResult = pinpad.initPinPad(PinPadTypeEnum.INTERNAL)
            if (pinPadInitResult != SdkResult.Success) {
                nexgoErrorLog("initPinPad returned $pinPadInitResult")
                eventEmitter.onNext(CardEvent.GeneralError {})
                cleanup()
            } else {
                // prompt for online or offline PIN
                val result = if (isOnlinePin) {
                    onlinePinRequested = true
                    nexgoDebugLog("cardPan= $cardPan")
                    pinpad.inputOnlinePin(
                            intArrayOf(4, 5, 6, 7, 8, 9, 10, 11, 12), // supported bin PIN lengths,
                            60,  // timeout in seconds
                            ByteUtils.string2ASCIIByteArray(cardPan), // PAN in ASCII encoding
                            keyIdService.mainKeyIndex, // specify the main key that was used to encrypt the working PIN key
                            PinAlgorithmModeEnum.ISO9564FMT3,
                            getPinListener(true)
                    )
                } else {
                    nexgoDebugLog("about to prompt for offline PIN")
                    pinpad.inputOfflinePin(
                            intArrayOf(4, 5, 6, 7, 8, 9, 10, 11, 12), // supported bin PIN lengths,
                            60, // timeout in seconds
                            getPinListener(false)
                    )
                }

                // check whether PIN prompt has been created
                if (result != SdkResult.Success) {
                    Timber.e("could not input PIN, result=$result, cardPan=$cardPan")
                    eventEmitter.onNext(CardEvent.PinEnteringFinished)
                    eventEmitter.onNext(CardEvent.GeneralError {})
                } else {
                    eventEmitter.onNext(CardEvent.PinEnteringStarted)
                    if (incorrectOfflinePinJustEntered) {
                        eventEmitter.onNext(CardEvent.WrongPin)
                        incorrectOfflinePinJustEntered = false
                    }
                }
            }
        }

        override fun onPrompt(prompt: PromptEnum?) {
            nexgoDebugLog("onPrompt: $prompt")
            // process PIN input result
            when (prompt) {
                PromptEnum.APP_SELECTION_IS_NOT_ACCEPTED -> {
                    Timber.e("onPrompt: ${PromptEnum.APP_SELECTION_IS_NOT_ACCEPTED.name} occurred")
                    eventEmitter.onNext(CardEvent.GeneralError {})
                }
                PromptEnum.OFFLINE_PIN_INCORRECT_TRY_AGAIN -> {
                    incorrectOfflinePinJustEntered = true
                    // further processed in the onFinish method
                }
                PromptEnum.OFFLINE_PIN_INCORRECT -> eventEmitter.onNext(CardEvent.WrongPinFinal)
                PromptEnum.OFFLINE_PIN_CORRECT -> {
                    // online PIN successful
                    Timber.d("onPrompt: OFFLINE_PIN_CORRECT")
                }
                else -> {
                    throw Exception("onPrompt is null")
                }
            }
            emvHandler.onSetPromptResponse(true)
        }

        override fun onRemoveCard() {
            nexgoDebugLog("onRemoveCard")
            emvHandler.onSetRemoveCardResponse()
            led.setLed(LightModeEnum.BLUE, false)
        }

        override fun onOnlineProc() {
            // the transaction must be authorized online...
            nexgoDebugLog("onOnlineProc")
            eventEmitter.onNext(
                    CardEvent.OnlineRequest(
                            data = CardData().apply {
                                // get request data
                                emvRequestData = "01" + getTagValuesConcatenated(emvRequestDataTags.toList(), emvHandler, EmvDataSourceEnum.FROM_KERNEL)
                                additionalEmvRequestData = "02" + getTagValuesConcatenated(emvRequestData2Tags.toList(), emvHandler, EmvDataSourceEnum.FROM_KERNEL)

                                presentationMethod = CardPresentation.CLESS

                                if (onlinePinRequested) {
                                    // convert hex representation to ascii, because the acquirer requires it
                                    val hexRepresentationInAscii = BytesUtil.byte2hex(submittedPinBlock).toByteArray()
                                    nexgoDebugLog("PIN submittedPinBlock=${BytesUtil.byte2hex(submittedPinBlock)}")
                                    nexgoDebugLog("PIN hexRepresentationInAscii=${BytesUtil.byte2hex(hexRepresentationInAscii)}")

                                    // uncomment following lines once Monet+ supports
                                    // the FMT3 PIN block format (that is used in the Nexgo SDK)
                                    // to send the PIN to the acquirer

                                    // pinBlock = hexRepresentationInAscii
                                    // addFlag(CardData.CardFlag.PIN_ENTERED)
                                }

                                // get from info entity or from TLV tag
                                // onConfirmCardInfo is sometimes skipped (Mastercard)
                                cardNumber = cardPan
                                track2 = cardTrack2
                                expirationDate = confirmedCardInfoEntity?.expiredDate
                                if (expirationDate == null) {
                                    track2?.let { track2 ->
                                        expirationDate = getExpiryDateFromTrack2(track2)
                                    }
                                }
                                nexgoDebugLog("emvHandler.signNeed: ${emvHandler.signNeed}")
                                if (emvHandler.signNeed) {
                                    addFlag(CardData.CardFlag.REQUIRES_SIGNATURE)
                                }
                            },
                            callback = { responseResult, acquirerResponseData ->

                                emvHandler.onSetOnlineProcResponse(if (responseResult?.isError == false) 0 else 1, EmvOnlineResultEntity().apply {
                                    this.authCode = acquirerResponseData?.approvalCode
                                    this.rejCode = responseResult?.responseCode
                                    this.recvField55 = BytesUtil.mergeBytes(
                                            acquirerResponseData?.issuerAuthenticationData?.let { TLV.fromData("91", it) }?.rawData,
                                            acquirerResponseData?.issuerScript
                                    )
                                    nexgoDebugLog("""
                                        isError = ${responseResult?.isError}
                                        authcode = $authCode
                                        rejCode = $rejCode
                                        recvField55 = ${BytesUtil.byte2hex(recvField55 ?: byteArrayOf(0x00))}
                                    """.trimIndent())
                                })
                            })
            )
        }

        private val cardPan: String?  get() = confirmedCardInfoEntity?.cardNo
                ?: getTagValuesConcatenated(listOf("5A"), emvHandler, EmvDataSourceEnum.FROM_KERNEL)

        private val cardTrack2: String? get() = confirmedCardInfoEntity?.tk2
                ?: getTagValuesConcatenated(listOf("57"), emvHandler, EmvDataSourceEnum.FROM_KERNEL)

        // get a list of concatenated tag values from the given source
        private fun getTagValuesConcatenated(tags: List<String>, emv: EmvHandler, source: EmvDataSourceEnum) = tags.mapNotNull { tag ->
            val value = emv.getTlv(ByteUtils.hexString2ByteArray(tag), source)
            if (value == null) {
                nexgoErrorLog("Tag: $tag not found")
                null
            } else {
                ByteUtils.byteArray2HexString(value).let { value ->
                    when (tag) {
                        // terminal country code and currency code must have the leading zero removed
                        TlvTags.terminalCountryCode, TlvTags.transactionCurrencyCode -> value.replaceFirst(regex = Regex("^0"), "")
                        else -> value
                    }
                }
            }
        }.joinToString(separator = "")

        override fun onReadCardAgain() {
            nexgoDebugLog("onReadCardAgain")
            eventEmitter.onNext(CardEvent.GeneralError {
                restartSearchForCard()
            })
        }

        override fun onAfterFinalSelectedApp() {
            // an AID has been selected
            nexgoDebugLog("onAfterFinalSelectedApp")
            if (initCardInfoEntity?.cardExistslot == CardSlotTypeEnum.RF) {
                val aid = emvHandler.getTlv(byteArrayOf(0x4F), EmvDataSourceEnum.FROM_KERNEL)
                val kernelId = emvHandler.getTlv(byteArrayOf(0x9F.toByte(), 0x2A.toByte()), EmvDataSourceEnum.FROM_KERNEL)
                if (kernelId != null) {
                    nexgoDebugLog("kernelId: ${BytesUtil.byte2hex(kernelId)}")
                }
                if (aid != null) {
                    val aid = ByteUtils.byteArray2HexString(aid).toUpperCase(Locale.ROOT)
                    val isMastercard = aid.startsWith("A000000004")
                    val isVisa = aid.startsWith("A000000003")
                    var kernelId: Byte? = null

                    if (isMastercard) {
                        nexgoDebugLog("onAfterFinalSelectedApp - Master Card selected (${aid})")
                        kernelId = KernelID.MASTER.toByte()

                        // set mastercard specific TLV tags regarding terminal capabilities and CVM
                        emvHandler.setTlv(byteArrayOf(0x9F.toByte(), 0x33.toByte()), byteArrayOf(0xE0.toByte(), 0x40.toByte(), 0xC8.toByte()))
                        emvHandler.setTlv(byteArrayOf(0xDF.toByte(), 0x81.toByte(), 0x18.toByte()), byteArrayOf(0x40.toByte()))
                        emvHandler.setTlv(byteArrayOf(0xDF.toByte(), 0x81.toByte(), 0x19.toByte()), byteArrayOf(0x40.toByte()))

                    } else if (isVisa) {
                        // visa selected
                        kernelId = KernelID.VISA.toByte()
                    }

                    kernelId?.let { kernelId ->
                        // add common TLV tags using the EmvParameterInitializer
                        val (config, record) = EmvParametersHelper(preferencesService).getTransactionConfigAndRecordForSearchCardData(searchCardInputData!!)
                        // some tags are already initialized by the Nexgo SDK,
                        // to not override them
                        val tagsToBeIgnored = listOf(
                                EmvTags.V_TAG_TM_9F66 // do not set transaction qualifiers
                        )

                        EmvParameterInitializer(config, record).initEmvParameters(
                                aid,
                                kernelId,
                                null,
                                CardType.RF,
                                com.stpos.a8pos.devicesdk.impl.mock.emv.MockEmvData(), // mock EMV data
                                searchCardInputData!!.sessionType, // transaction type
                                CardData(), // empty card data
                                preferencesService // preferences service
                        )
                                .asList()
                                .forEach {
                                    // check whether the tag should be ignored
                                    if (it.tag !in tagsToBeIgnored) {
                                        nexgoDebugLog("TLV: ${BytesUtil.byte2hex(BytesUtil.hexString2ByteArray(it.tag))} : ${BytesUtil.byte2hex(it.bytesValue)}")
                                        emvHandler.setTlv(BytesUtil.hexString2ByteArray(it.tag), it.bytesValue)
                                    } else {
                                        nexgoDebugLog("Tag ${it.tag} skipped, value: ${BytesUtil.byte2hex(it.bytesValue)}")
                                    }
                                }
                    }
                }
            }

            // set the merchant name (tag 9F4E)
            val merchantName = preferencesService.merchantName
            val name: ByteArray = try {
                merchantName.toByteArray(charset("GBK"))
            } catch (e: UnsupportedEncodingException) {
                merchantName.toByteArray()
            }
            emvHandler.setTlv(BytesUtil.hexString2ByteArray(EmvTags.EMV_TAG_TM_MCHNAMELOC), name)

            emvHandler.onSetAfterFinalSelectedAppResponse(true)
        }

        override fun onFinish(result: Int, resultEntity: EmvProcessResultEntity?) {
            nexgoDebugLog("onFinish: result=$result")
            eventEmitter.onNext(CardEvent.Processing)
            led.setLed(LightModeEnum.BLUE, false)
            // process the EMV process result
            when (result) {
                SdkResult.Fail -> {
                    eventEmitter.onNext(CardEvent.CardPresentError(preferredAction = CardEvent.CardPresentError.Action.TRY_AGAIN))
                    restartSearchForCardDelayed()
                }
                SdkResult.Emv_Declined, SdkResult.Emv_Offline_Declined -> {
                    eventEmitter.onNext(CardEvent.CardNotAccepted(CardEvent.CardNotAccepted.Reason.EMV_DECLINED, null) {
                        restartSearchForCard()
                    })
                }
                SdkResult.Success -> {
                    if ((initCardInfoEntity?.cardExistslot == CardSlotTypeEnum.RF || initCardInfoEntity?.cardExistslot == CardSlotTypeEnum.ICC1)) {
                        // check whether the card was inserted in supported slots
                        nexgoDebugLog("Kernel Success: CardSlotTypeEnum.RF")
                        eventEmitter.onNext(CardEvent.KernelVerificationSuccessful)
                    }
                    cleanup()
                }
                SdkResult.Emv_Candidatelist_Empty -> {
                    // try different card, as the candidate list is empty
                    eventEmitter.onNext(CardEvent.CardReadErrorTryDifferent(
                            reason = CardEvent.CardReadErrorTryDifferent.Reason.USE_CHIP_MAGSTRIPE_OR_DIFFERENT_CARD
                    ) {
                        restartSearchForCard()
                    })
                }
                SdkResult.Emv_FallBack -> {
                    eventEmitter.onNext(CardEvent.UseFallback {
                        restartSearchForCard()
                    })
                }
                SdkResult.Emv_Plz_See_Phone -> {
                    nexgoDebugLog("SdkResult.Emv_Plz_See_Phone")
                    eventEmitter.onNext(CardEvent.CdCvmNeeded {
                        restartSearchForCard()
                    })
                }
                SdkResult.Emv_Cancel,
                SdkResult.Emv_Terminate -> {
                    // EMV has been cancelled/terminated
                    eventEmitter.onNext(CardEvent.EmvCanceled)
                    stopEmvAndSearchingForCard()
                }
                else -> {
                    stopEmvAndSearchingForCard()
                    eventEmitter.onNext(CardEvent.CardPresentError(preferredAction = CardEvent.CardPresentError.Action.TRY_AGAIN))
                    restartSearchForCardDelayed()
                }
            }
        }

        override fun onSelApp(appNameList: MutableList<String>?, appInfoList: MutableList<CandidateAppInfoEntity>?, isFirstSelect: Boolean) {
            // the candidate list contains multiple items, the cardholder must selected the AID to be used
            nexgoDebugLog("onSelApp")
            // appNameList => names of available applications
            // appInfoList => further information about of available applications
            // isFirstSelect => whether the selection is being made for the first time
            // -------------------------------------------------
            // this method must call the emvHandler.onSetSelAppResponse, where
            // the first parameter is the index of the selected application
            // (starting at 1!)

            if (appNameList != null && appInfoList != null) {
                val mappedAidList: List<CandidateAID> = appNameList.zip(appInfoList).map { (displayName, info) ->
                    CandidateAID().apply {
                        aid = info.aid
                        appLabel = info.appLabel ?: displayName.toByteArray()
                        langPref = info.langPrefer ?: "cs".toByteArray()
                        issCTIndex = info.icti
                    }
                }

                eventEmitter.onNext(
                        CardEvent.SelectAid(mappedAidList) { selectedAid ->
                            onAidSelectedByCardholder(selectedAid, mappedAidList, appInfoList)
                        }
                )
            }
        }

        // unused right now
        override fun onRequestAmount() = TODO("On request amount")
        override fun onConfirmEcSwitch() = TODO("onConfirmEcSwitch")
        override fun onCertVerify(p0: String?, p1: String?) = TODO("onCertVerify")

        private fun onAidSelectedByCardholder(
                selectedAid: CandidateAID,
                mappedAidList: List<CandidateAID>,
                appInfoList: MutableList<CandidateAppInfoEntity>) {
            // find the corresponding AID
            val correspondingAid = appInfoList.indexOfFirst {
                it.aid.contentEquals(selectedAid.aid)
            }
            if (correspondingAid >= 0) {
                // corresponding AID found, increment 1 to the index,
                // because the Nexgo SDK requires it
                emvHandler.onSetSelAppResponse(
                        correspondingAid + 1
                )
            } else {
                // some internal error occurred as the corresponding
                // AID was not found
                Timber.e(
                        """The selected AID ($selectedAid) from the SelectAid callback has not been found in the original list of AIDs.
                                            | appInfoList = [${appInfoList.joinToString { it.aid.toString() }}]
                                            | mappedAidList = [${mappedAidList.joinToString { it.aid.toString() }}]
                                        """.trimMargin()
                )
                eventEmitter.onNext(CardEvent.GeneralError {
                    restartSearchForCard()
                })
            }
        }
    }
}